#include "iodefine.h"
#include "machine.h"
#include "ctype.h"

#define	printf	((int (*)(const char *,...))0x00007c7c)
#define	scanf	((int (*)(const char *,...))0x00007cb8)

#define	SD_CD		(PE.DR.BIT.B11)
#define	SD_CS		(PE.DR.BIT.B9)

// ------------------------------------------------------------
struct disk_t {
	unsigned int First_sect_LBA;
	unsigned int BPB_BytesPerSec;
	unsigned int BPB_SecPerClus;
	unsigned int BPB_RsvdSecCnt;
	unsigned int BPB_NumFATs;
	unsigned int BPB_RootEntCnt;
	unsigned int BPB_FATSz16;
	unsigned int First_RDE_sect;
	unsigned int First_FAT_sect;
	unsigned int First_Data_sect;
} DISK;


struct file_t {
	unsigned char Filename[12];
	unsigned int n;					// RDE ([琔) n Ԗڂ̃t@C
	unsigned int attr;				// 
	unsigned int File_year;			// N
	unsigned int File_month;		// 
	unsigned int File_date;			// 
	unsigned int File_hour;			// 
	unsigned int File_min;			// 
	unsigned int File_sec;			// b
	unsigned int FstClusLO;
	unsigned int FileSize;
	unsigned char *Data;			// t@C{
};


// ------------------------------------------------------------
// -- O[oϐ --
unsigned char dt[512];			// ZN^[hp[N
unsigned char fat[1024];		// FAT
unsigned char rde[512];			// RDE

struct file_t File;			// t@C
unsigned char FileData0[4096];		// t@C{


// ------------------------------------------------------------
// ------------------------------------------------------------
void init_CMT0()
{
	STB.CR4.BIT._CMT = 0;

	CMT0.CMCNT = 0;
	CMT0.CMCSR.BIT.CKS = 1;		// CKSݒ(0:1/8, 1:1/32, 2:1/128, 3:1/512)
}


// ------------------------------------------------------------
void wait_us(unsigned int us) {
	unsigned int val;

	val = us * 10 / 16;
	if (val >= 0xffff)
		val = 0xffff;

	CMT0.CMCOR = val;
	CMT0.CMCSR.BIT.CMF &= 0;
	CMT.CMSTR.BIT.STR0 = 1;

	while (CMT0.CMCSR.BIT.CMF == 0)
		;
	CMT0.CMCSR.BIT.CMF = 0;
	CMT.CMSTR.BIT.STR0 = 0;
}


// ------------------------------------------------------------
void init_SCI2()
{
	STB.CR3.BIT._SCI2 = 0;
	SCI2.SCSCR.BYTE= 0x00;
	SCI2.SCSMR.BIT.CA = 1;		// NbN
	SCI2.SCBRR = 12;			// 384kbps
	SCI2.SCSDCR.BIT.DIR = 1;	// MSB first
	wait_us(1);
	PFC.PECRL3.BIT.PE10MD = 2;	// PE10 .. TxD
	PFC.PECRL3.BIT.PE8MD = 2;	// PE8 .. SCK
	PFC.PECRL2.BIT.PE7MD = 2;	// PE7 .. RxD
	PFC.PEIORL.BIT.B12 = 0;		// PE12 .. WP
	PFC.PEIORL.BIT.B11 = 0;		// PE11 .. CD
	PFC.PEIORL.BIT.B9 = 1;		// PE9 .. CSo
	PE.DR.BIT.B9 = 1;			// CSl
	SCI2.SCSCR.BYTE |= 0x30;	// TE, RE = 1
}


// ------------------------------------------------------------
unsigned char SPI_tx_rx(unsigned char ch)
{
	while (!SCI2.SCSSR.BIT.TDRE)
		;
	SCI2.SCTDR = ch;
	SCI2.SCSSR.BIT.TDRE = 0;

	while (!SCI2.SCSSR.BIT.RDRF)
		;
	ch = SCI2.SCRDR;
	SCI2.SCSSR.BIT.RDRF = 0;
	return (ch);
}


// ------------------------------------------------------------
int card_exist()
{
	return (PE.DR.BIT.B11 ? 0 : 1);
}


// ------------------------------------------------------------
unsigned char calc_CRC7(unsigned char *data, int len)
{
	int i, j;
	char crc, dt;

	crc = 0;
	for (i = 0; i < len; i++) {
		dt = *data++;
		for (j = 0; j < 8; j++) {
			crc <<=1;
			if ((crc & 0x80) ^ (dt & 0x80))
				crc ^= 0x09;
			dt <<= 1;
		}
	}
	return (crc & 0x7f);
}


// ------------------------------------------------------------
unsigned char SD_send_cmd(unsigned char cmd, int arg)
{
	int i;
	unsigned char cmd_token[6], ret;

	while (1) {
		if (SPI_tx_rx(0xff) == 0xff)
			break;
	}

	cmd_token[0] = cmd;
	cmd_token[1] = (arg >> 24) & 0xff;
	cmd_token[2] = (arg >> 16) & 0xff;
	cmd_token[3] = (arg >> 8) & 0xff;
	cmd_token[4] = arg & 0xff;
	cmd_token[5] = (calc_CRC7(cmd_token, 5) << 1) | 0x01;

	for (i = 0; i < 6; i++)
		SPI_tx_rx(cmd_token[i]);

	do {
		ret = SPI_tx_rx(0xff);
	} while (ret & 0x80);

	return (ret);
}


// ------------------------------------------------------------
int Enter_SPI_mode()
{
	int i;
	unsigned char ret;

	SD_CS = 1;				// CS negate

	for (i = 0; i < 10; i++)
		SPI_tx_rx(0xff);

	SD_CS = 0;				// CS assert

	ret = SD_send_cmd(0x40, 0);
	if (ret != 0x01)
		return (-1);

	do {
		ret = SD_send_cmd(0x41, 0);
	} while (ret != 0x00);

	return (1);
}


// ------------------------------------------------------------
void read_sector(unsigned int sect, unsigned char *dt)
{
	unsigned char ch;
	int i;

	ch = SD_send_cmd(0x51, sect * 512);
	while (ch != 0xfe)
		ch = SPI_tx_rx(0xff);

	for (i = 0; i < 512; i++)
		*dt++ = SPI_tx_rx(0xff);
	SPI_tx_rx(0xff);				// CRC
	SPI_tx_rx(0xff);				// CRC
}


// ------------------------------------------------------------
void print_sector(unsigned int sect, unsigned char *dt)
{
	int i, j;
	unsigned int adr;

	adr = sect * 512;

	printf("           ");
	for (i = 0; i < 16; i++) {
		printf("+%x ", i);
		if (i == 7)
			printf(" ");
	}
	printf("\n");

	for (i = 0; i < 32; i++) {
		printf("%08x : ", adr);
		for (j = 0; j < 16; j++) {
			printf("%02x ", dt[i * 16 + j]);
			if (j == 7)
				printf(" ");
		}
		printf(" ");
		for (j = 0; j < 16; j++) {
			if (isprint(dt[i * 16 + j]))
				printf("%c", dt[i * 16 + j]);
			else
				printf(".");
			if (j == 7)
				printf(" ");
		}
		printf("\n");
		adr += 16;
	}

}


// ------------------------------------------------------------
// ------------------------------------------------------------
// MBR ͂
//
void ex_MBR()
{
	read_sector(0, dt);
	DISK.First_sect_LBA = 0;
}



// ------------------------------------------------------------
// BPB ͂
//
void ex_BPB()
{
	read_sector(DISK.First_sect_LBA, dt);
	DISK.BPB_BytesPerSec = 0;
	DISK.BPB_SecPerClus = 0;
	DISK.BPB_RsvdSecCnt = 0;
	DISK.BPB_NumFATs = 0;
	DISK.BPB_RootEntCnt = 0;
	DISK.BPB_FATSz16 = 0;
}


// ------------------------------------------------------------
// RDE 擾
//
void get_RDE()
{
	DISK.First_RDE_sect = 0;
	read_sector(DISK.First_RDE_sect, rde);
}


// ------------------------------------------------------------
// RDE  n Ԗڂ̃t@C擾
//
void get_file_info(struct file_t *file)
{
	int i;
	unsigned int offset;

	offset = file->n * 32;

	// -- t@C --
	for (i = 0; i < 8; i++)
		file->Filename[i] = rde[offset + i];

	file->Filename[8] = '.';

	// -- gq --
	for (i = 8; i < 11; i++)
		file->Filename[i + 1] = rde[offset + i];

	file->attr = 0;
	file->File_hour = 0;
	file->File_min = 0;
	file->File_sec = 0;
	file->File_year = 0;
	file->File_month = 0;
	file->File_date = 0;
	file->FstClusLO = 0;
	file->FileSize = 0;
}


// ------------------------------------------------------------
// FAT 擾
// TCY傫t@CɑΉ邽߁AFAT  2 ZN^擾Ă
//
void get_FAT()
{
	DISK.First_FAT_sect = 0;
	read_sector(DISK.First_FAT_sect, fat);
	read_sector(DISK.First_FAT_sect + 1, &fat[512]);
}


// ------------------------------------------------------------
// NX^`F[\
//
void ex_FAT(struct file_t *file)
{
	unsigned int cluster;

	cluster = 0;
	printf("%d ", cluster);
	while (cluster != 0xffff) {
		cluster = 0;
		printf(" -> %d ", cluster);
	}
	printf("\n");
}


// ------------------------------------------------------------
// t@C擾
//
void get_File(struct file_t *file)
{
	int i, j;
	unsigned int sz;
	unsigned int cluster;

	DISK.First_Data_sect = 0;
	cluster = 0;
	sz = 0;
	while (cluster != 0xffff) {
		cluster = 0;
	}
}


// ------------------------------------------------------------
// t@Ce\
//
void print_File(struct file_t *file)
{
	int i;

	for (i = 0; i < file->FileSize; i++)
		printf("%c", file->Data[i]);
	printf("\n");
}


// ------------------------------------------------------------
// ZN^_v
//
void dump_SECT()
{
	unsigned int SECT_NR;

	printf("\nSector number = 0x");
	scanf("%x", &SECT_NR);
	read_sector(SECT_NR, dt);
	print_sector(SECT_NR, dt);
}


// ------------------------------------------------------------
// ------------------------------------------------------------
void main()
{
	int i;

	init_CMT0();
	init_SCI2();

	if (!card_exist())
		printf("No card found\n");
	else {
		if (Enter_SPI_mode() < 0)
			printf("SPI mode Err\n");
		else {
			printf("SPI mode\n");

			SD_send_cmd(0x50, 512);		// ubNTCY=512


			// ------------------------------------------------------------
			// MBR  .. First_sect_LBA 擾
//			ex_MBR();
//			printf("1st_sector(LBA) = 0x%x\n\n", DISK.First_sect_LBA); 


			// ------------------------------------------------------------
			// BPB  .. BPB_RsvdSecCnt 擾
//			ex_BPB();
//			printf("BPB_BytesPerSec = %d\n", DISK.BPB_BytesPerSec);
//			printf("BPB_SecPerClus = %d\n", DISK.BPB_SecPerClus);
//			printf("BPB_RsvdSecCnt = %d\n", DISK.BPB_RsvdSecCnt);
//			printf("BPB_NumFATs = %d\n", DISK.BPB_NumFATs);
//			printf("BPB_RootEntCnt = %d\n", DISK.BPB_RootEntCnt);
//			printf("BPB_FATSz16 = %d\n\n", DISK.BPB_FATSz16);


			// ------------------------------------------------------------
			// FAT 擾 -> O[oϐ fat[] Ɋi[
//			get_FAT();


			// ------------------------------------------------------------
			// RDE 擾 -> O[oϐ rde[] Ɋi[
//			get_RDE();
//			printf("First RDE sect = 0x%x\n", DISK.First_RDE_sect);


			// ------------------------------------------------------------
			// RDE t@C擾
//			File.n = 3;				// RDE 3 Ԗڂ̃t@C
//			File.Data = FileData0;
//			get_file_info(&File);
//			for (i = 0; i < 12; i++)
//				printf("%c", File.Filename[i]);
//			printf("\n");
//			printf(" = 0x%x\n", File.attr);
//			printf("Time = %02d:%02d:%02d\n", File.File_hour, File.File_min, File.File_sec);
//			printf("Date = %4d/%02d/%02d\n", File.File_year, File.File_month, File.File_date);
//			printf("FileSize = %d\n\n", File.FileSize);
//			printf("FstClusLO = %d\n", File.FstClusLO);


			// ------------------------------------------------------------
			// File ɑΉ FAT ͂ANX^`F[\
//			ex_FAT(&File);


			// ------------------------------------------------------------
			// File ɑΉt@C{̂擾 -> File.data[] Ɋi[
//			get_File(&File);


			// ------------------------------------------------------------
			// t@Ce\
//			printf("FileSize = %d\n", File.FileSize);
//			print_File(&File);


			// ------------------------------------------------------------
			while (1) {
				dump_SECT();		// ZN^_v
			}
		}
	}

	SD_CS = 1;				// CS negate

	while (1)
		;
}